def _black(text):
    return f"\033[90m{text}\033[0m"


def _red(text):
    return f"\033[91m{text}\033[0m"


def _green(text):
    return f"\033[92m{text}\033[0m"


def _yellow(text):
    return f"\033[93m{text}\033[0m"


def _blue(text):
    return f"\033[94m{text}\033[0m"


def _magenta(text):
    return f"\033[95m{text}\033[0m"


def _cyan(text):
    return f"\033[96m{text}\033[0m"


def _white(text):
    return f"\033[97m{text}\033[0m"


def _bold_black(text):
    return f"\033[1;90m{text}\033[0m"


def _bold_red(text):
    return f"\033[1;91m{text}\033[0m"


def _bold_green(text):
    return f"\033[1;92m{text}\033[0m"


def _bold_yellow(text):
    return f"\033[1;93m{text}\033[0m"


def _bold_blue(text):
    return f"\033[1;94m{text}\033[0m"


def _bold_magenta(text):
    return f"\033[1;95m{text}\033[0m"


def _bold_cyan(text):
    return f"\033[1;96m{text}\033[0m"


def _bold_white(text):
    return f"\033[1;97m{text}\033[0m"


def _dim_black(text):
    return f"\033[30m{text}\033[0m"


def _dim_red(text):
    return f"\033[31m{text}\033[0m"


def _dim_green(text):
    return f"\033[32m{text}\033[0m"


def _dim_yellow(text):
    return f"\033[33m{text}\033[0m"


def _dim_blue(text):
    return f"\033[34m{text}\033[0m"


def _dim_magenta(text):
    return f"\033[35m{text}\033[0m"


def _dim_cyan(text):
    return f"\033[36m{text}\033[0m"


def _dim_white(text):
    return f"\033[37m{text}\033[0m"

